"""Yet Another Pixel Classifier.

Usage:
  yapic train <network> <image_path> <label_path> [options]
  yapic predict <network> <image_path> <output_path> [options]
  yapic deploy <network> <image_path> <output_path> [options]

  <network>               Either a model file in h5 format to use a pretrained
                          model or specific string to initialize a new model.
                          Choose 'unet_2d' or 'unet_multi_z' to initialize a
                          new model. Use path/to/my/pretrained_model.h5 to
                          continue training of a pretrained keras model
                          (train mode), to apply a trained model to a set of
                          images (predict mode), or to export a model to
                          DeepImageJ (deploy mode).
  <image_path>            Path to image files. You can use wildcards, e.g.
                          "my_data/*.tif". In deploy mode, define one specific
                          tif image file (as example image).
  <label_path>            Path to label files. Either tif or ilp (Ilastik
                          project file).
                          Wildcards are supported.
                          Examples: "path/to/my_ilastik_project.ilp",
                                    "path/to/my_label_images/*.tif"
  <output_path>           Path to directory to store prediction resuts
                          (prediction mode) or exported model (deploy mode).

Options:
  -n --normalize=NORM     Set pixel normalization scope [default: local]
                          For minibatch-wise normalization choose
                          'local_z_score' or 'local'. For global normalization
                          use global_<min>+<max> (e.g. 'global_0+255' for 8-bit
                          images and 'global_0+65535' for 16-bit images)
                          Choose 'off' to deactivate.
  --cpu                   Use the CPU (not recommended for training).
  --gpu=VISIBLE_DEVICES   If you wanrt to use specific gpus. To use gpu 0,
                          set '0'. To use gpus 2 and 3, set '2,3'
  -h --help               Show this screen.
  --version               Show version.

Train Options:
  -e --epochs=MAX_EPOCHS  Maximum number of epochs to train [default: 5000].
  -a --augment=AUGMENT    Set augmentation method for training [default: flip]
                          Choose 'flip' and/or 'rotate' and/or 'shear'.
                          Use '+' to specify multiple augmentations
                          (e.g. flip+rotate).
  -v --valfraction=VAL    Fraction of images to be used for validation
                          [default: 0.2].
  -f --file=CLASSIFER     Path to trained model [default: model.h5].
  --steps=STEPS           Steps per epoch [default: 50].
  --equalize              Equalize label weights to promote less frequent
                          labels.
  --csvfile=LOSSDATA      Path to csv file for training loss data
                          [default: loss.csv].

Predict Options:
  --split                 Save prediction images of different classes into
                          separate images. By default, predictions are saved
                          as ImageJ multichannel images.

Deploy Options:
  -s --size=MODELSIZE     Size of the network to be exported. Large networks
                          are applied faster in DeepImageJ, but consume more
                          RAM. There are three options:
                          'small' (112 x 112 pixels),
                          'middle' (224 x 224 pixels),
                          'large' (368 x 368 pixels) [default: middle]
  --skip-predict          Skip computation of prediction image.
  --author=AUTHOR         Name of the model's authors [default: n/a]
  --url=URL               Url to model publication [default: http://]
  --credit=CREDIT         [default: n/a]
  --mdversion=MDVERSION   Model version [default: n/a]
  --reference=REFERENCE   Publication reference of model [default: n/a]



"""
from docopt import docopt

import os
import sys
import h5py
from yapic.session import Session
from yapic.deepimagej_exporter import DeepimagejExporter
from yapic.version import __version__ as package_version

import logging
logger = logging.getLogger(__name__)


def main(args):

    if args['--cpu']:
        # deactivate gpu for tensorflow
        os.environ['CUDA_VISIBLE_DEVICES'] = ''
    if args['--gpu']:
        # define gpu hardware
        os.environ['CUDA_VISIBLE_DEVICES'] = args['--gpu']

    image_path = os.path.abspath(os.path.expanduser(args['<image_path>']))
    model_name = args['<network>']

    if args['deploy']:
        output_path = args['<output_path>']
        predict_example_image = not args['--skip-predict']

        assert os.path.isfile(model_name)

        exp = DeepimagejExporter(model_name,
                                 output_path,
                                 image_path)
        exp.export_as_deepimagej(
            author=args['--author'],
            version=args['--version'],
            url=args['--url'],
            credit=args['--credit'],
            reference=args['--reference'],
            size=args['--size'],
            applymodel=predict_example_image)
        return

    norm_string = args['--normalize']
    s = Session()

    if args['train']:

        label_path = os.path.abspath(os.path.expanduser(args['<label_path>']))
        aug_string = args['--augment']
        max_epochs = int(args['--epochs'])
        steps_per_epoch = int(args['--steps'])
        log_filename = args['--csvfile']
        model_export_filename = args['--file']
        valfraction = float(args['--valfraction'])

        s.load_training_data(image_path, label_path)

        models_available = ['unet_2d',
                            'unet_multi_z',
                            'convnet_for_unittest']

        if os.path.isfile(model_name):
            s.load_model(model_name)
        elif model_name in models_available:
            size_xy = 572
            if model_name == 'unet_2d' or model_name == 'convnet_for_unittest':
                size_z = 1
            if model_name == 'unet_multi_z':
                size_z = 5
            if model_name == 'convnet_for_unittest':
                size_xy = 100

            s.make_model(model_name, (size_z, size_xy, size_xy))

        s.set_normalization(norm_string)
        s.set_augmentation(aug_string)

        if valfraction > 0:
            s.define_validation_data(valfraction)

        s.train(max_epochs=max_epochs,
                steps_per_epoch=steps_per_epoch,
                log_filename=log_filename,
                model_filename=model_export_filename)
        
        f = h5py.File(model_export_filename, 'r+')
        lbl_map_list = [[key, val] for key, val in s.lbl_map.items()]
        f.create_dataset('lbl_map', data=lbl_map_list)

    if args['predict']:
        output_path = args['<output_path>']
        assert os.path.isfile(model_name), '<network> must be a h5 model file'

        multichannel = not(args['--split'])

        s.load_prediction_data(image_path, output_path)
        s.load_model(model_name)
        s.set_normalization(norm_string)
        s.predict(multichannel=multichannel)


def entry_point():
    arguments = docopt(__doc__, version=f"YAPiC {package_version}")
    res = main(arguments)
    sys.exit(res)


if __name__ == '__main__':
    entry_point()



from yapic.network import make_model, load_keras_model
import yapic.utils as ut
import numpy as np
from yapic_io.training_batch import TrainingBatch
from yapic_io.prediction_batch import PredictionBatch
from yapic_io.connector import io_connector
from yapic_io.dataset import Dataset
import tensorflow.keras as keras
import sys
from tensorflow.python.framework.tensor_shape import Dimension

import logging
logger = logging.getLogger(__name__)


class Session(object):
    '''
    A session is used for training a model with a connected dataset (pixels
    and labels) or for predicting connected data (pixels) with an already
    trained model.




    Parameters
    ----------
    data : yapic_io.TrainingBatch
        Connector object for binding pixel and label data
    '''

    def __init__(self):

        self.dataset = None
        
        self.lbl_map = None
        
        self.model = None
        self.data = None
        self.data_val = None
        self.history = None
        self.data_predict = None
        self.log_filename = None

        self.output_tile_size_zxy = None
        self.padding_zxy = None

    def load_training_data(self, image_path, label_path):
        '''
        Connect to a training dataset.

        Parameters
        ----------
        image_path : string
            Path to folder with tiff images.
        label_path : string
            Path to folder with label tiff images or path to ilastik project
            file (.ilp file).
        '''
        
        _connector = io_connector(image_path, label_path)
        self.lbl_map = _connector.labelvalue_mapping[0]
        
        self.dataset = Dataset(_connector)

        msg = '\n\nImport taining dataset:\n{}\n'.format(
            self.dataset.pixel_connector.__repr__())
        sys.stdout.write(msg)

    def load_prediction_data(self, image_path, save_path):
        '''
        Connect to a prediction dataset.

        Parameters
        ----------
        image_path : string
            Path to folder with tiff images to predict.
        save_path : string
            Path to folder for saving prediction images.
        '''

        self.dataset = Dataset(io_connector(image_path,
                                            '/tmp/this_should_not_exist',
                                            savepath=save_path))
        msg = '\n\nImport dataset for prediction:\n{}\n'.format(
            self.dataset.pixel_connector.__repr__())
        sys.stdout.write(msg)

    def make_model(self, model_name, input_tile_size_zxy):
        '''
        Initialzes a neural network and sets tile sizes of the data connector
        accordingly to model input/output shapes.

        Parameters
        ----------
        model_name : string
            Either 'unet_2d' or 'unet_2p5d'
        input_tile_size_zxy: (nr_zslices, nr_x, nr_y)
            Input shape of the model. Large input shapes require large memory
            for used GPU hardware. For 'unet_2d', nr_zslices has to be 1.
        '''

        sys.stdout.write('\n\nInitialize model {}\n'.format(model_name))
        assert len(input_tile_size_zxy) == 3
        nr_channels = self.dataset.image_dimensions(0)[0]
        input_size_czxy = [nr_channels] + list(input_tile_size_zxy)
        n_classes = len(self.dataset.label_values())

        self.model = make_model(model_name, n_classes, input_size_czxy)

        output_tile_size_zxy = self.model.output_shape[-4:-1]

        output_tile_size_zxy = [v.value
                                if isinstance(v, Dimension) else v
                                for v in output_tile_size_zxy]

        self._configure_minibatch_data(input_tile_size_zxy,
                                       output_tile_size_zxy)

    def load_model(self, model_filepath):
        '''
        Import a Keras model in hfd5 format.

        Parameters
        ----------
        model_filepath : string
            Path to .h5 model file
        '''

        model = load_keras_model(model_filepath)

        n_classes_model = model.output_shape[-1]
        output_tile_size_zxy = model.output_shape[-4:-1]
        n_channels_model = model.input_shape[-1]
        input_tile_size_zxy = model.input_shape[-4:-1]

        n_classes_data = len(self.dataset.label_values())
        n_channels_data = self.dataset.image_dimensions(0)[0]

        msg = ('nr of model classes ({}) and data classes ({}) '
               'is not equal').format(n_classes_model, n_classes_data)
        if n_classes_data > 0:
            assert n_classes_data == n_classes_model, msg

        msg = ('nr of model channels ({}) and iamge channels ({}) '
               'is not equal').format(n_channels_model, n_channels_data)
        assert n_channels_data == n_channels_model, msg

        self._configure_minibatch_data(input_tile_size_zxy,
                                       output_tile_size_zxy)
        self.model = model

    def _configure_minibatch_data(self,
                                  input_tile_size_zxy,
                                  output_tile_size_zxy):
        padding_zxy = tuple(((np.array(input_tile_size_zxy) -
                            np.array(output_tile_size_zxy))/2).astype(np.int))

        self.data_val = None
        self.data = TrainingBatch(self.dataset,
                                  output_tile_size_zxy,
                                  padding_zxy=padding_zxy)
        self.data.set_normalize_mode('local')
        self.data.set_pixel_dimension_order('bzxyc')
        next(self.data)
        self.output_tile_size_zxy = output_tile_size_zxy
        self.padding_zxy = padding_zxy

    def define_validation_data(self, valfraction):
        '''
        Splits the dataset into a training fraction and a validation fraction.

        Parameters
        ----------
        valfraction : float
            Approximate fraction of validation data. Has to be between 0 and 1.
        '''
        msg = ('\nConfiuring validation dataset '
               '({} validation data, {} training data)').format(
                   valfraction,
                   1 - valfraction)
        sys.stdout.write(msg)
        if self.data_val is not None:
            logger.warning('skipping define_validation_data: already defined')
            return None

        if self.data is None:
            logger.warning('skipping, data not defined yet')
            return None
        self.data.remove_unlabeled_tiles()
        self.data_val = self.data.split(valfraction)

    def train(self, max_epochs=3000, steps_per_epoch=24, log_filename=None,
              model_filename='model.h5'):
        '''
        Starts a training run.

        Parameters
        ----------
        max_epochs : int
            Number of epochs.
        steps_per_epoch : int
            Number of training steps per epoch.
        log_filename : string
           Path to the csv file for logging loss and accuracy.
        model_filename : string
           Path to h5 keras model file


        Notes
        -----
        Validation is executed once each epoch.
        Logging to csv file is executed once each epoch.
        '''

        callbacks = []

        if self.data_val:
            save_model_callback = keras.callbacks.ModelCheckpoint(
                                        model_filename,
                                        monitor='val_loss',
                                        verbose=0,
                                        save_best_only=True)
        else:
            save_model_callback = keras.callbacks.ModelCheckpoint(
                                        model_filename,
                                        monitor='loss',
                                        verbose=0,
                                        save_best_only=True)
        callbacks.append(save_model_callback)

        if log_filename:
            callbacks.append(keras.callbacks.CSVLogger(log_filename,
                                                       separator=',',
                                                       append=False))

        training_data = ((mb.pixels(), mb.weights())
                         for mb in self.data)

        validation_steps_per_epoch = steps_per_epoch
        if self.data_val:
            validation_data = ((mb.pixels(), mb.weights())
                               for mb in self.data_val)
        else:
            validation_data = None
            validation_steps_per_epoch = None

        self.history = self.model.fit_generator(
                            training_data,
                            validation_data=validation_data,
                            epochs=max_epochs,
                            validation_steps=validation_steps_per_epoch,
                            steps_per_epoch=steps_per_epoch,
                            callbacks=callbacks,
                            workers=0)

        return self.history

    def predict(self, multichannel=False):
        data_predict = PredictionBatch(self.dataset,
                                       2,
                                       self.output_tile_size_zxy,
                                       self.padding_zxy)
        data_predict.set_normalize_mode('local')
        data_predict.set_pixel_dimension_order('bzxyc')

        if multichannel:
            data_predict.multichannel_output_on()
        else:
            data_predict.multichannel_output_off()
        print('multichannel output: {}'.format(data_predict.multichannel))

        for item_nr, item in enumerate(data_predict):
            msg = ('Writing probability map tile'
                   ' {} of {}...\n'.format(item_nr+1,
                                           len(data_predict)))
            sys.stdout.write(msg)
            sys.stdout.flush()
            result = self.model.predict(item.pixels())
            item.put_probmap_data(result)

        sys.stdout.write('Writing probability maps finished.\n')

    def set_augmentation(self, augment_string):
        '''
        Define data augmentation settings for model training.

        Parameters
        ----------
        augment_string : string
            Choose 'flip' and/or 'rotate' and/or 'shear'.
            Use '+' to specify multiple augmentations (e.g. flip+rotate).
        '''

        if self.data is None:
            logger.warning(
                'could not set augmentation to {}. Run make_model() first')
            return

        ut.handle_augmentation_setting(self.data, augment_string)

    def set_normalization(self, norm_string):
        '''
        Set pixel normalization scope.

        Parameters
        ----------
        norm_string : string
            For minibatch-wise normalization choose 'local_z_score' or 'local'.
            For global normalization use global_<min>+<max>
            (e.g. 'global_0+255' for 8-bit images and 'global_0+65535' for
            16-bit images).
            Choose 'off' to deactivate.
        '''

        if self.data is None:
            logger.warning(
                'could not set normalizarion to {}. Run make_model() first')
            return

        ut.handle_normalization_setting(self.data, norm_string)



from unittest import TestCase
import yapic.network as netw
import os

os.environ['CUDA_VISIBLE_DEVICES'] = ''


class TestNetwork(TestCase):

    def test_setup_unet_2d(self):

        n_classes = 2
        size_czxy = (3, 1, 572, 572)

        model = netw.setup_network('unet_2d', n_classes, size_czxy)

        self.assertEqual(model.input_shape, (None, 1, 572, 572, 3))
        self.assertEqual(model.output_shape, (None, 1, 358, 358, 2))

    def test_setup_unet_multi_z(self):

        n_classes = 2
        size_czxy = (3, 5, 572, 572)

        model = netw.setup_network('unet_multi_z', n_classes, size_czxy)

        self.assertEqual(model.input_shape, (None, 5, 572, 572, 3))
        self.assertEqual(model.output_shape, (None, 1, 354, 354, 2))

    def test_setup_convnet_for_unittest(self):

        n_classes = 2
        size_czxy = (3, 1, 150, 150)

        model = netw.setup_network('convnet_for_unittest',
                                   n_classes,
                                   size_czxy)

        self.assertEqual(model.input_shape, (None, 1, 150, 150, 3))
        self.assertEqual(model.output_shape, (None, 1, 138, 138, 2))




from tensorflow.keras import optimizers, losses
from tensorflow.keras import backend as K
from tensorflow.keras import metrics
from tensorflow import keras
import logging


logger = logging.getLogger(__name__)


def setup_network(network_name, N_classes, input_size_czxy):
    '''
    Builds up the network named `network_name` with `N_channels` input channels
    and `N_class_list` output classes.
    '''
    module_path = 'yapic.networks.{}'.format(network_name)
    try:
        root_mod = __import__(module_path)
    except ImportError:
        msg = 'Could not import network "{name}" (file networks/{name}.py)'
        raise ImportError(msg.format(name=network_name))

    network_module = getattr(root_mod.networks, network_name)
    network = network_module.build_network(N_classes, input_size_czxy)

    return network


def count_labels(y):
    return K.sum(K.cast(K.any(K.not_equal(y, 0), axis=-1), dtype=K.floatx()))


def correct_mean(y):
    return K.prod(K.cast(K.shape(y)[:-1], dtype=K.floatx())) / count_labels(y)


def corr_categ_crsentropy(y_true, y_pred):
    return losses.categorical_crossentropy(y_true,
                                           y_pred) * correct_mean(y_true)


def compile_model(network, learning_rate=1e-3, momentum=0.9):
    '''
    Compiles the network
    '''

    def accuracy(y_true, y_pred):
        return metrics.categorical_accuracy(y_true,
                                            y_pred) * correct_mean(y_true)

    optimize = optimizers.SGD(lr=learning_rate,
                              momentum=momentum,
                              nesterov=True)
    network.compile(optimizer=optimize,
                    loss=corr_categ_crsentropy,
                    metrics=[accuracy])

    return network


def make_model(network_name, N_classes, input_size_czxy,
               learning_rate=1e-3, momentum=0.9):
    logger.info(
        'Building network "{}" with input size {} (czxy) and {} classes...'
        .format(network_name, input_size_czxy, N_classes))
    net = setup_network(network_name, N_classes, input_size_czxy)

    net = compile_model(net, learning_rate=learning_rate, momentum=momentum)
    logger.debug('Network built with output size {}.'
                 .format(net.output_shape))
    return net


def load_keras_model(filepath):

    try:
        model = keras.models.load_model(
            filepath,
            custom_objects={'corr_categ_crsentropy': corr_categ_crsentropy})
    except ValueError:
            model = keras.models.load_model(
                filepath,
                custom_objects={
                    'corrected_categorical_crossentropy':
                        corr_categ_crsentropy})

    return model



from tensorflow import keras
import logging
import os
import math

logger = logging.getLogger(os.path.basename(__file__))


def convolve(net, n_filters, filter_size, padding):
    '''
    double convolution step
    '''
    net = keras.layers.Conv2D(n_filters,
                              filter_size,
                              padding=padding,
                              activation='relu',
                              kernel_initializer='glorot_normal',
                              data_format='channels_last')(net)
    net = keras.layers.Conv2D(n_filters,
                              filter_size,
                              padding=padding,
                              activation='relu',
                              kernel_initializer='glorot_normal',
                              data_format='channels_last')(net)
    return net


def contract(net, n_filters, padding):
    net = convolve(net, n_filters, (3, 3), padding)
    to_concat = net

    net = keras.layers.MaxPooling2D(pool_size=(2, 2),
                                    data_format='channels_last')(net)
    return net, to_concat


def expand(net, to_concat, n_filters, padding):
    net = keras.layers.UpSampling2D(size=(2, 2),
                                    data_format='channels_last')(net)
    net = convolve(net, n_filters, (2, 2), padding)

    shape_diff = to_concat.shape[-2] - net.shape[-2]
    cropping = int(math.floor(int(shape_diff)/2.))

    cropped = keras.layers.Cropping2D(cropping=cropping,
                                      data_format='channels_last')(to_concat)

    net = keras.layers.Concatenate(axis=-1)([net, cropped])
    net = convolve(net, n_filters, (3, 3), padding)

    return net


def build_network(N_classes, input_size_czxy, squeeze=False, padding='valid'):
    '''
    Builds the original U-Net by Ronneberger et al. (2015) for 2d images
    '''
    logger.debug('Building network with {} classes'.format(N_classes))

    assert len(input_size_czxy) == 4
    N_channels = input_size_czxy[0]
    size_z = input_size_czxy[1]
    size_xy = input_size_czxy[3]

    assert N_classes >= 2
    assert N_channels >= 1
    assert size_z == 1
    assert input_size_czxy[2] == input_size_czxy[3]

    if squeeze:
        net = keras.layers.Input(shape=(size_xy, size_xy, N_channels))
        input_net = net
    else:
        net = keras.layers.Input(shape=(size_z, size_xy, size_xy, N_channels))
        input_net = net
        net = keras.layers.Reshape((size_xy, size_xy, N_channels))(net)

    net, to_concat_1 = contract(net, 64, padding)
    net, to_concat_2 = contract(net, 128, padding)
    net, to_concat_3 = contract(net, 256, padding)
    net, to_concat_4 = contract(net, 512, padding)
    net, to_concat_5 = contract(net, 1024, padding)
    net = expand(to_concat_5, to_concat_4, 512, padding)
    net = expand(net, to_concat_3, 256, padding)
    net = expand(net, to_concat_2, 128, padding)
    net = expand(net, to_concat_1, 64, padding)

    net = keras.layers.Conv2D(N_classes,
                              (1, 1),
                              activation='softmax',
                              data_format='channels_last')(net)

    size_xy_out = int(net.shape[-2])
    if not squeeze:
        net = keras.layers.Reshape((1,
                                    size_xy_out,
                                    size_xy_out,
                                    N_classes))(net)
    model = keras.models.Model(inputs=input_net, outputs=net, name='unet_2d')

    return model



import logging
import re

logger = logging.getLogger(__name__)


def handle_augmentation_setting(mbatch, augment):

    mbatch.augment_by_flipping(False)
    mbatch.augment_by_rotation(False)
    mbatch.augment_by_shear(False)

    augment = set(augment.split('+'))

    if 'flip' in augment:
        mbatch.augment_by_flipping(True)
        logger.info('activate augmentation mode: flip')

    if 'rotate' in augment:
        mbatch.augment_by_rotation(True)
        logger.info('activate augmentation mode: rotate')

    if 'shear' in augment:
        mbatch.augment_by_shear(True)
        logger.info('activate augmentation mode: shear')

    if not mbatch.augmentation:
        logger.info('Augmentation is not active')

    undefined = augment - {'flip', 'rotate', 'shear'}
    if undefined:
        raise ValueError(
            'Incorrect augmentation mode(s): {}'.format(undefined))


def handle_normalization_setting(mbatch, normalize_str):
    if normalize_str in ['local', 'local_z_score', 'off']:
        mbatch.set_normalize_mode(normalize_str)
        logger.info('set normalization mode to {}'.format(normalize_str))
        return

    pattern = r'(global)_(-?\d{1,10})\+(-?\d{1,10})'
    regex_match = re.match(pattern, normalize_str)

    assert regex_match, 'Incorrect normalize mode: {}'.format(normalize_str)

    mode, min, max = regex_match.groups()
    minmax = (int(min), int(max))
    assert minmax[0] < minmax[1], 'max larger than min: {}'.format(minmax)
    mbatch.set_normalize_mode(mode, minmax=minmax)



from yapic.networks import unet_2d
from tensorflow import keras


def build_network(N_classes, input_size_czxy):

    N_channels = input_size_czxy[0]
    size_z = input_size_czxy[1]
    size_xy = input_size_czxy[3]

    assert size_z == 5

    input_size_z_layer = list(input_size_czxy)
    input_size_z_layer[1] = 1

    input = keras.layers.Input(shape=(size_z, size_xy, size_xy, N_channels))

    # one 2d unet for each z-slice
    z_layer = unet_2d.build_network(N_classes,
                                    (N_channels,
                                     1,
                                     size_xy,
                                     size_xy))
    models_single_z_layer = []
    for crop_z in [(0, 4), (1, 3), (2, 2), (3, 1), (4, 0)]:
        input_single_z_layer = keras.layers.Cropping3D(
                                        cropping=(crop_z, (0, 0), (0, 0)),
                                        data_format='channels_last')(input)
        models_single_z_layer.append(z_layer(input_single_z_layer))

    net = keras.layers.Concatenate(axis=1)(models_single_z_layer)

    net = keras.layers.Conv3D(25,
                              (3, 3, 3),
                              activation='relu',
                              kernel_initializer='glorot_normal',
                              data_format='channels_last')(net)

    net = keras.layers.Conv3D(N_classes,
                              (3, 3, 3),
                              activation='softmax',
                              kernel_initializer='glorot_normal',
                              data_format='channels_last')(net)

    model = keras.models.Model(inputs=input, outputs=net)

    return model



from unittest import TestCase
from yapic.deepimagej_exporter import DeepimagejExporter
from yapic.session import Session
import os
import shutil
import tensorflow as tf
import pytest

base_path = os.path.dirname(__file__)


def train_test_model_unet_2d_1channel_3classes():
    img_path = os.path.join(
        base_path,
        '../test_data/shapes/pixels/*')
    label_path = os.path.join(
        base_path,
        '../test_data/shapes/labels.ilp')
    savepath = os.path.join(
        base_path,
        '../test_data/tmp')
    model_filename = os.path.join(savepath,
                                  'model_unet_2d_1channel_3classes.h5')
    if os.path.isfile(model_filename):
        return model_filename

    os.makedirs(savepath, exist_ok=True)

    t = Session()
    t.load_training_data(img_path, label_path)
    t.make_model('unet_2d', (1, 220, 220))
    t.train(max_epochs=2,
            steps_per_epoch=2,
            log_filename=os.path.join(savepath, 'log.csv'),
            model_filename=model_filename)

    return model_filename


def train_test_model_unet_2d_1channel_2classes():
    img_path = os.path.join(
        base_path,
        '../test_data/shapes/pixels/*')
    label_path = os.path.join(
        base_path,
        '../test_data/shapes/labels_2classes.ilp')
    savepath = os.path.join(
        base_path,
        '../test_data/tmp')
    model_filename = os.path.join(savepath,
                                  'model_unet_2d_1channel_2classes.h5')
    if os.path.isfile(model_filename):
        return model_filename

    os.makedirs(savepath, exist_ok=True)

    t = Session()
    t.load_training_data(img_path, label_path)
    t.make_model('unet_2d', (1, 220, 220))
    t.train(max_epochs=2,
            steps_per_epoch=2,
            log_filename=os.path.join(savepath, 'log.csv'),
            model_filename=model_filename)

    return model_filename


def train_test_model_unet_2d_3channels_2classes():
    img_path = os.path.join(
        base_path,
        '../test_data/shapes/pixels_rgb/*')
    label_path = os.path.join(
        base_path,
        '../test_data/shapes/labels_2classes.ilp')
    savepath = os.path.join(
        base_path,
        '../test_data/tmp')
    model_filename = os.path.join(savepath,
                                  'model_unet_2d_3channels_2classes.h5')
    if os.path.isfile(model_filename):
        return model_filename

    os.makedirs(savepath, exist_ok=True)

    t = Session()
    t.load_training_data(img_path, label_path)
    t.make_model('unet_2d', (1, 220, 220))
    t.train(max_epochs=2,
            steps_per_epoch=2,
            log_filename=os.path.join(savepath, 'log.csv'),
            model_filename=model_filename)

    return model_filename


def train_test_model_convnet():
    img_path = os.path.join(
        base_path,
        '../test_data/shapes/pixels/*')
    label_path = os.path.join(
        base_path,
        '../test_data/shapes/labels.ilp')
    savepath = os.path.join(
        base_path,
        '../test_data/tmp')
    model_filename = os.path.join(savepath, 'model_convnet.h5')
    if os.path.isfile(model_filename):
        return model_filename

    os.makedirs(savepath, exist_ok=True)

    t = Session()
    t.load_training_data(img_path, label_path)
    t.make_model('convnet_for_unittest', (1, 100, 100))
    t.train(max_epochs=2,
            steps_per_epoch=2,
            log_filename=os.path.join(savepath, 'log.csv'),
            model_filename=model_filename)

    return model_filename


class TestDeepimagejExporter(TestCase):

    def setUp(self):

        model_path = train_test_model_unet_2d_1channel_3classes()
        print('saved unet_2d as {}'.format(model_path))

        model_path = train_test_model_unet_2d_1channel_2classes()
        print('saved unet_2d as {}'.format(model_path))

        model_path = train_test_model_unet_2d_3channels_2classes()
        print('saved unet_2d as {}'.format(model_path))

    def test_is_model_unet_2d(self):

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))

        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model'))

        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_3classes.h5')
        print('model_path: {}'.format(model_path))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)
        assert exp._is_model_unet_2d()

        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_2classes.h5')
        print('model_path: {}'.format(model_path))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)
        assert exp._is_model_unet_2d()

        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_3channels_2classes.h5')
        print('model_path: {}'.format(model_path))
        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels_rgb/pixels_1.tif'))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)
        assert exp._is_model_unet_2d()

    @pytest.mark.slow
    def test_reshape_unet_2d_1channel_3classes(self):

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))

        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model'))

        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_3classes.h5')

        print('model_path: {}'.format(model_path))

        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)

        exp._reshape_unet_2d(size='small')
        assert exp.model_reshaped.input_shape == (None, 112, 112, 1)
        assert exp.model_reshaped.output_shape == (None, 112, 112, 3)

    @pytest.mark.slow
    def test_reshape_unet_2d_1channel_2classes(self):

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))

        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model'))

        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_2classes.h5')

        print('model_path: {}'.format(model_path))

        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)

        exp._reshape_unet_2d(size='small')
        assert exp.model_reshaped.input_shape == (None, 112, 112, 1)
        assert exp.model_reshaped.output_shape == (None, 112, 112, 2)

    @pytest.mark.slow
    def test_reshape_unet_2d_3channels_2classes(self):

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels_rgb/pixels_1.tif'))

        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model'))

        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_3channels_2classes.h5')

        print('model_path: {}'.format(model_path))

        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)

        exp._reshape_unet_2d(size='small')
        assert exp.model_reshaped.input_shape == (None, 112, 112, 3)
        assert exp.model_reshaped.output_shape == (None, 112, 112, 2)

    @pytest.mark.slow
    def test_export_as_tensorflow_model_1(self):
        # 1 channel 3 classes

        tf_version = [int(num) for num in tf.__version__.split('.')]

        if tf_version[0] != 1:
            # deepimagej supports only tensorflow version 1
            return

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))
        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model_1channel_3classes'))
        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_3classes.h5')
        print('model_path: {}'.format(model_path))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)
        exp._reshape_unet_2d(size='small')
        exp._export_as_tensorflow_model()

    @pytest.mark.slow
    def test_export_as_tensorflow_model_2(self):
        # 1 channel 2 classes

        tf_version = [int(num) for num in tf.__version__.split('.')]

        if tf_version[0] != 1:
            # deepimagej supports only tensorflow version 1
            return

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))
        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model_1channel_2classes'))
        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_2classes.h5')
        print('model_path: {}'.format(model_path))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)
        exp._reshape_unet_2d(size='small')
        exp._export_as_tensorflow_model()

    @pytest.mark.slow
    def test_export_as_tensorflow_model_3(self):
        # 3 channels 2 classes

        tf_version = [int(num) for num in tf.__version__.split('.')]

        if tf_version[0] != 1:
            # deepimagej supports only tensorflow version 1
            return

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels_rgb/pixels_1.tif'))
        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model_3channels_2classes'))
        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_3channels_2classes.h5')
        print('model_path: {}'.format(model_path))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)
        exp._reshape_unet_2d(size='small')
        exp._export_as_tensorflow_model()

    @pytest.mark.slow
    def test_update_metadata(self):

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))
        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model_1channel_2classes'))
        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_2classes.h5')
        print('model_path: {}'.format(model_path))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)
        exp._reshape_unet_2d(size='small')

        exp._update_metadata(author='Some Name')
        print(exp.metadata)
        assert exp.metadata['name'] == 'exported_model_1channel_2classes'
        assert exp.metadata['input_tensor_dimensions'] == (-1, 112, 112, 1)
        assert exp.metadata['patch_size'] == 112
        assert exp.metadata['author'] == 'Some Name'

    def test_format_xml(self):

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))
        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/exported_model_cpl'))
        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_2classes.h5')
        print('model_path: {}'.format(model_path))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)
        exp._reshape_unet_2d(size='small')

        exp._update_metadata(author='Some Name',
                             version='1.0.0',
                             url='https://my-site.org',
                             credit='some other names',
                             reference='Name et al. 2020'
                             )

        os.makedirs(save_path, exist_ok=True)
        exp._format_xml()

    @pytest.mark.slow
    def test_export_as_deepimagej(self):
        # 1 channel 2 classes

        tf_version = [int(num) for num in tf.__version__.split('.')]

        if tf_version[0] != 1:
            # deepimagej supports only tensorflow version 1
            return

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))
        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/my_packaged_u_net'))
        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_2classes.h5')
        print('model_path: {}'.format(model_path))

        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)

        shutil.rmtree(save_path, ignore_errors=True)
        exp.export_as_deepimagej(
            author='Some Name',
            version='1.0.0',
            url='https://my-site.org',
            credit='some other names',
            reference='Name et al. 2020',
            size='small')

    def test_export_as_deepimagej_without_apply(self):
        # 1 channel 2 classes

        tf_version = [int(num) for num in tf.__version__.split('.')]

        if tf_version[0] != 1:
            # deepimagej supports only tensorflow version 1
            return

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels/pixels_1.tiff'))
        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/my_packaged_u_net'))
        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_2classes.h5')
        print('model_path: {}'.format(model_path))

        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)

        shutil.rmtree(save_path, ignore_errors=True)
        exp.export_as_deepimagej(
            author='Some Name',
            version='1.0.0',
            url='https://my-site.org',
            credit='some other names',
            reference='Name et al. 2020',
            size='small',
            applymodel=False)


    @pytest.mark.slow
    def test_apply_model(self):

        example_image_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/shapes/pixels_small/pixels_1.tif'))
        save_path = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/my_packaged_u_net'))
        temp_dir = os.path.abspath(os.path.join(
            base_path,
            '../test_data/tmp/temp_dir'))
        model_path = os.path.join(
            base_path,
            '../test_data/tmp/model_unet_2d_1channel_2classes.h5')
        print('model_path: {}'.format(model_path))
        exp = DeepimagejExporter(model_path,
                                 save_path,
                                 example_image_path)

        os.makedirs(save_path, exist_ok=True)
        os.makedirs(temp_dir, exist_ok=True)
        exp.apply_model(normalization_mode='local')



from unittest import TestCase
from yapic.session import Session
import os
import skimage
import numpy as np
import pytest
import shutil

base_path = os.path.dirname(__file__)
os.environ['CUDA_VISIBLE_DEVICES'] = ''


class TestSessionMethods(TestCase):

    def test_set_augmentation(self):

        img_path = os.path.join(
            base_path,
            '../test_data/shapes/pixels/*')
        label_path = os.path.join(
            base_path,
            '../test_data/shapes/labels.ilp')

        t = Session()
        t.load_training_data(img_path, label_path)
        t.make_model('unet_2d', (1, 572, 572))

        t.set_augmentation('flip+rotate')
        self.assertEqual(t.data.augmentation, {'rotate', 'flip'})
        t.set_augmentation('flip+rotate+shear')
        self.assertEqual(t.data.augmentation, {'rotate', 'flip', 'shear'})
        t.set_augmentation('flip')
        self.assertEqual(t.data.augmentation, {'flip'})

    def test_set_normalization(self):

        img_path = os.path.join(
            base_path,
            '../test_data/shapes/pixels/*')
        label_path = os.path.join(
            base_path,
            '../test_data/shapes/labels.ilp')

        t = Session()
        t.load_training_data(img_path, label_path)
        t.make_model('unet_2d', (1, 572, 572))

        t.set_normalization('local')
        assert t.data.normalize_mode == 'local'
        assert t.data.global_norm_minmax is None

        t.set_normalization('global_0+255')
        assert t.data.normalize_mode == 'global'
        assert t.data.global_norm_minmax == (0, 255)

    def test_set_normalization_if_no_data(self):

        t = Session()
        # t.load_training_data(img_path, label_path)

        t.set_normalization('local')

    def test_load_model(self):

        os.environ['CUDA_VISIBLE_DEVICES'] = ''
        img_path = os.path.join(
            base_path,
            '../test_data/shapes/pixels/*')
        label_path = os.path.join(
            base_path,
            '../test_data/shapes/labels.ilp')
        savepath = os.path.join(
            base_path,
            '../test_data/tmp')

        os.makedirs(savepath, exist_ok=True)

        # train initial model and save it
        t = Session()
        t.load_training_data(img_path, label_path)
        t.make_model('convnet_for_unittest', (1, 100, 100))
        t.train(max_epochs=3,
                steps_per_epoch=2,
                log_filename=os.path.join(savepath, 'log.csv'),
                model_filename=os.path.join(savepath, 'model.h5'))

        # initialize new sassion and load pretrained model
        t2 = Session()
        t2.load_training_data(img_path, label_path)

        assert t2.model is None
        t2.load_model(os.path.join(savepath, 'model.h5'))
        assert t2.model is not None

        # train model further
        t2.train(max_epochs=3,
                 steps_per_epoch=2,
                 log_filename=os.path.join(savepath, 'log.csv'),
                 model_filename=os.path.join(savepath, 'model_continued.h5'))


class TestEnd2End(TestCase):

    def test_shape_data_fast(self):

        # train a classifier and predict training data
        os.environ['CUDA_VISIBLE_DEVICES'] = '2'

        img_path = os.path.join(
            base_path,
            '../test_data/shapes/pixels/*')
        label_path = os.path.join(
            base_path,
            '../test_data/shapes/labels.ilp')
        savepath = os.path.join(
            base_path,
            '../test_data/tmp')

        os.makedirs(savepath, exist_ok=True)

        t = Session()
        t.load_training_data(img_path, label_path)
        t.make_model('convnet_for_unittest', (1, 100, 100))

        t.train(max_epochs=3,
                steps_per_epoch=4,
                log_filename=os.path.join(savepath, 'log.csv'),
                model_filename=os.path.join(savepath, 'model.h5'))
        t.load_prediction_data(img_path, savepath)
        t.predict()

        artifacts = os.listdir(savepath)
        assert 'log.csv' in artifacts
        assert 'pixels_1_class_1.tif' in artifacts
        assert 'pixels_1_class_2.tif' in artifacts
        assert 'pixels_1_class_3.tif' in artifacts
        assert 'pixels_2_class_1.tif' in artifacts
        assert 'pixels_2_class_2.tif' in artifacts
        assert 'pixels_2_class_3.tif' in artifacts

        shutil.rmtree(savepath)

    @pytest.mark.slow
    def test_shape_data(self):

        # train a classifier and predict training data
        os.environ['CUDA_VISIBLE_DEVICES'] = '2'

        img_path = os.path.join(
            base_path,
            '../test_data/shapes/pixels/*')
        label_path = os.path.join(
            base_path,
            '../test_data/shapes/labels.ilp')
        savepath = os.path.join(
            base_path,
            '../test_data/tmp')

        os.makedirs(savepath, exist_ok=True)

        t = Session()
        t.load_training_data(img_path, label_path)

        t.make_model('convnet_for_unittest', (1, 100, 100))

        t.train(max_epochs=50,
                steps_per_epoch=50,
                log_filename=os.path.join(savepath, 'log.csv'),
                model_filename=os.path.join(savepath, 'model.h5'))
        t.load_prediction_data(img_path, savepath)
        t.predict()

        # read prediction images and compare with validation data
        def read_images(image_nr, class_nr):
            if class_nr == 1:
                shape = 'circles'
            if class_nr == 2:
                shape = 'triangles'

            filename = os.path.join(
                                savepath,
                                'pixels_{}_class_{}.tif'.format(image_nr,
                                                                class_nr))
            print(filename)
            prediction_img = np.squeeze(skimage.io.imread(filename))
            filename = os.path.join(
                                    savepath,
                                    '../shapes/val/{}_{}.tiff'.format(
                                        shape,
                                        image_nr))
            print(filename)
            val_img = np.squeeze(skimage.io.imread(filename))
            return prediction_img, val_img

        prediction_img, val_img = read_images(1, 1)
        accuracy = np.mean(prediction_img[val_img > 0][:])
        print(accuracy)
        self.assertTrue(accuracy > 0.6)

        prediction_img, val_img = read_images(1, 2)
        accuracy = np.mean(prediction_img[val_img > 0][:])
        print(accuracy)
        self.assertTrue(accuracy > 0.6)

        prediction_img, val_img = read_images(2, 1)
        accuracy = np.mean(prediction_img[val_img > 0][:])
        print(accuracy)
        self.assertTrue(accuracy > 0.6)

        prediction_img, val_img = read_images(2, 2)
        accuracy = np.mean(prediction_img[val_img > 0][:])
        print(accuracy)
        self.assertTrue(accuracy > 0.6)

        shutil.rmtree(savepath)



import os
from setuptools import setup, find_packages
from yapic.version import __version__

try:
    import tensorflow as tf
except ModuleNotFoundError:
    msg = ('You have to install tensorflow or tensorflow-gpu version '
           '1.12, 1.13, 1.14, 1.15 or 2.1'
           'before installing YAPiC')
    raise ModuleNotFoundError(msg)


reqs = ['yapic_io>=0.1.2',
        'docopt>=0.6.2',
        'numpy>=1.15.4']


def readme():
    README_md = os.path.join(os.path.dirname(__file__), 'README.md')
    with open(README_md) as f:
        return f.read()


ns = os.environ.get('CI_PROJECT_NAMESPACE', 'idaf')

setup(name='yapic',
      version=__version__,
      description='Yet another Pixel Classifier (based on deep learning)',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://yapic.github.io/yapic/',
      author='Manuel Schoelling, Christoph Moehl',
      author_email=('manuel.schoelling@gmx.de, '
                    'christoph.oliver.moehl@gmail.com'),
      packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
      include_package_data=True,
      zip_safe=False,
      install_requires=reqs,
      test_suite='nose.collector',
      entry_points={
          'console_scripts': ['yapic=yapic.main:entry_point'],
      },
      tests_require=['nose', 'coverage', 'nose-timer', 'nose-deadline'])



from tensorflow import keras


def build_network(N_classes, input_size_czxy):
    '''
    builds a small model suitable for running on cpu
    '''
    assert len(input_size_czxy) == 4

    N_channels = input_size_czxy[0]
    size_z = input_size_czxy[1]
    size_xy = input_size_czxy[3]

    assert N_classes >= 2
    assert N_channels >= 1
    assert size_z == 1
    assert input_size_czxy[2] == input_size_czxy[3]

    net = keras.layers.Input(shape=(size_z, size_xy, size_xy, N_channels))
    input_net = net

    net = keras.layers.Reshape((size_xy, size_xy, N_channels))(net)

    n_convolutions = 6
    n_filters = 50
    filter_size = (3, 3)
    for _ in range(n_convolutions):
        net = keras.layers.Conv2D(n_filters,
                                  filter_size,
                                  activation='relu',
                                  kernel_initializer='glorot_normal',
                                  data_format='channels_last')(net)

    net = keras.layers.Conv2D(N_classes,
                              (1, 1),
                              activation='softmax',
                              data_format='channels_last')(net)

    size_xy_out = int(net.shape[-2])

    net = keras.layers.Reshape((1, size_xy_out, size_xy_out, N_classes))(net)
    model = keras.models.Model(inputs=input_net, outputs=net)

    return model



from yapic.session import Session
from yapic.networks import unet_2d
import os
import tensorflow as tf
from tensorflow.keras import backend as K
import time
import xml.etree.ElementTree as ET
import shutil
# from tensorflow.keras import backend as K


class DeepimagejExporter(object):
    '''
    A DeepImageJ exporter provides methods to deploy Keras models
    trained with YAPiC to the ImageJ plugin DeepImageJ.

    Parameters
    ----------
    model_path : string
        Path to Keras model in h5 format.
    save_path : string
        Path to directory where the exported model and metadata is saved.
    example_image_path: string
        Path to example input image in tif format.
    '''

    def __init__(self, model_path, save_path, example_image_path):

        save_dir = os.path.dirname(save_path)
        if len(save_dir) == 0:
            save_dir = './'
        msg = '{} does not exist'.format(save_dir)
        assert os.path.isdir(save_dir), msg
        self.save_path = save_path

        self.s = Session()
        self.s.load_prediction_data(example_image_path, self.save_path)
        self.s.load_model(model_path)

        self.model_path = model_path
        self.example_image_path = example_image_path

        msg = 'model is not unet_2d, cannot be exported to deepimagej'
        assert self._is_model_unet_2d(), msg

        self.model_reshaped = None
        self.metadata = None

        self.template_dir = os.path.join(
            os.path.dirname(__file__),
            'templates/deepimagej101')

    def export_as_deepimagej(self,
                             author='n/a',
                             url='http://',
                             credit='n.a',
                             version='n.a',
                             reference='n/a',
                             size='small',
                             applymodel=True):

        self._reshape_unet_2d(size=size)
        self._update_metadata(author=author,
                              version=version,
                              url=url,
                              credit=credit,
                              reference=reference)
        self._export_as_tensorflow_model()
        self._format_xml()

        shutil.copyfile(os.path.join(self.template_dir, 'postprocessing.txt'),
                        os.path.join(self.save_path, 'postprocessing.txt'))
        shutil.copyfile(os.path.join(self.template_dir, 'preprocessing.txt'),
                        os.path.join(self.save_path, 'preprocessing.txt'))

        if applymodel:
            self.apply_model('local')
        shutil.copyfile(self.example_image_path,
                        os.path.join(self.save_path, 'exampleImage.tiff'))

    def apply_model(self, normalization_mode):

        self.s.set_normalization(normalization_mode)
        self.s.predict(multichannel=True)

        result_img_name = os.path.basename(self.example_image_path)
        save_path = os.path.join(self.save_path,
                                 result_img_name)
        new_save_path = save_path.replace(result_img_name, 'resultImage.tiff')
        os.rename(save_path, new_save_path)

    def _is_model_unet_2d(self):
        return self.s.model.name == 'unet_2d'
        # return self.s.model.count_params() == 32424323

    def _reshape_unet_2d(self, size='middle'):

        if size == 'small':
            shape_xy = 112
        elif size == 'middle':
            shape_xy = 224
        elif size == 'large':
            shape_xy = 368
        else:
            shape_xy = 112

        print('reshape to {}'.format(shape_xy))
        N_classes = self.s.model.output_shape[-1]
        N_channels = self.s.model.input_shape[-1]

        self.model_reshaped = unet_2d.build_network(
                             N_classes,
                             (N_channels, 1, shape_xy, shape_xy),
                             squeeze=True,
                             padding='same')

        self.model_reshaped.set_weights(self.s.model.get_weights())

    def _update_metadata(self,
                         author='n/a',
                         url='http://',
                         credit='n.a',
                         version='n.a',
                         reference='n/a'):

        if self.model_reshaped is None:
            return
        if self.metadata is None:
            self.metadata = {}

        self.metadata['name'] = os.path.basename(self.save_path)
        self.metadata['author'] = author
        self.metadata['url'] = url
        self.metadata['credit'] = credit
        self.metadata['version'] = version
        self.metadata['reference'] = reference

        date_format = '%a %b %d %H:%M:%S %Z %Y'
        self.metadata['date'] = time.strftime(date_format, time.localtime())

        N_channels = self.model_reshaped.input_shape[-1]
        size_xy = self.model_reshaped.input_shape[2]

        self.metadata['channels'] = N_channels
        self.metadata['input_tensor_dimensions'] = (-1,
                                                    size_xy,
                                                    size_xy,
                                                    N_channels)
        self.metadata['patch_size'] = size_xy
        self.metadata['padding'] = int(size_xy * 0.19)

        # metadata = {'name': 'my_model',
        #             'author': 'n/a',
        #             'url': 'http://',
        #             'credit': 'n/a',
        #             'version': 'n/a',
        #             'reference': 'n/a',
        #             'date': 'Tue Mar 31 17:18:06 CEST 2020',
        #             'test_image_size_xy': (512, 512),
        #             'input_tensor_dimensions': (-1, 112, 112, 3),
        #             'patch_size': (112),
        #             'padding': 10}

    def _format_xml(self):

        if self.metadata is None:
            return

        xml_path = os.path.join(
            self.template_dir,
            'config.xml')

        tree = ET.parse(xml_path)

        key_mapping = (
                       (('ModelInformation', 'Name'),
                        'name'),
                       (('ModelInformation', 'Author'),
                        'author'),
                       (('ModelInformation', 'URL'),
                        'url'),
                       (('ModelInformation', 'Credit'),
                        'credit'),
                       (('ModelInformation', 'Version'),
                        'version'),
                       (('ModelInformation', 'Date'),
                        'date'),
                       (('ModelInformation', 'Reference'),
                        'reference'),
                       (('ModelCharacteristics', 'Channels'),
                        'channels'),
                       (('ModelCharacteristics', 'InputTensorDimensions'),
                        'input_tensor_dimensions'),
                       (('ModelCharacteristics', 'PatchSize'),
                        'patch_size'),
                       (('ModelCharacteristics', 'Padding'),
                        'padding'),
                       )

        for item in key_mapping:
            value = str(self.metadata[item[1]])
            if item[1] == 'input_tensor_dimensions':
                value = value.replace('(', ',')\
                             .replace(')', ',')\
                             .replace(' ', '')
            tree.find(item[0][0]).find(item[0][1]).text = value

        save_path = os.path.join(self.save_path, 'config.xml')
        tree.write(save_path)

    def _export_as_tensorflow_model(self):

        model = self.model_reshaped
        builder = tf.saved_model.builder.SavedModelBuilder(self.save_path)

        signature = tf.saved_model.signature_def_utils.predict_signature_def(
                        inputs={'input':  model.input},
                        outputs={'output': model.output})

        signature_def_map = {
            tf.saved_model
              .signature_constants
              .DEFAULT_SERVING_SIGNATURE_DEF_KEY: signature}

        builder.add_meta_graph_and_variables(
            K.get_session(),
            [tf.saved_model.tag_constants.SERVING],
            signature_def_map=signature_def_map)
        builder.save()
